#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level :mod:`QMessageBox`-based error handling facilities.

See Also
----------
:mod:`betsee.util.widget.guimessage`
    Error-agnostic :class:`QMessageBox` facilities.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable exceptions on application startup, the
# top-level of this module may import *ONLY* from submodules guaranteed to:
# * Exist, including standard Python and BETSEE modules. This does *NOT*
#   include BETSE modules, which are *NOT* guaranteed to exist at this point.
#   For simplicity, however, all core PySide2 submodules are assumed to exist.
# * Never raise exceptions on importation (e.g., due to module-level logic).
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

import re, sys, traceback
from PySide2.QtWidgets import QMessageBox
from betsee.util.app import guiapp

# ....................{ INSTALLERS                        }....................
def install_exception_hook() -> None:
    '''
    Install a global exception hook overriding :mod:`PySide2`'s default insane
    exception handling behaviour with sane exception handling.

    By default, :mod:`PySide2`:

    #. Catches **uncaught exceptions** (i.e., exceptions automatically
       propagated up the call stack without being caught) raised during the
       GUI's event loop processing.
    #. Prints the tracebacks for these exceptions to standard error.
    #. Ignores these exceptions by silently returning control back to the main
       event handling loop.

    This behaviour is entirely invisible to end users and hence insane. This
    function addresses this by installing a new handler both interactively
    displaying *and* non-interactively logging exceptions.

    Caveats
    ----------
    Ideally, this function should be called *before* entering this event loop
    (i.e., calling the :meth:`betsee.util.app.guiapp.GUI_APP._exec` method).
    '''

    # Default insane global exception hook, preserved in case of catastrophe.
    default_exception_handler = sys.excepthook

    # Custom sane global exception hook implemented as a closure requiring this
    # default insane global exception hook as a fallback.
    def exception_hook(exception_type, exception, exception_traceback):
        '''
        Application-specific global exception hook.

        See Also
        ----------
        :func:`install_handler`
            Further details.
        '''

        # Attempt to...
        try:
            # Import from BETSE. Since this hook should only ever be installed
            # *AFTER* BETSE is validated to be importable, this should succeed.
            from betse.util.io.log import logs

            # Log this exception.
            logs.log_exception(exception)

            # Display a PySide2-based message box displaying this exception.
            show_exception(exception)
        # If this exception handling itself raises an exception...
        except Exception as exception_exception:
            # Defer to the default global exception hook, presumably printing
            # both this new and the prior exception to stderr.
            default_exception_handler(
                type(exception_exception),
                exception_exception,
                traceback.extract_stack(exception_exception))

            # Exit the current application with failure status.
            sys.exit(1)

    # Replace the default global exception hook with this handler.
    sys.excepthook = exception_hook

# ....................{ SHOWERS                           }....................
def show_error(
    # Mandatory parameters.
    title: str,
    synopsis: str,

    # Optional parameters.
    exegesis: str = None,
    details: str = None,
) -> None:
    '''
    Display the passed error message(s) as a :mod:`QMessageBox`-driven modal
    message box of the :class:`QApplication` singleton for this application.

    Caveats
    ----------
    This function necessarily instantiates and initializes this singleton if
    needed. Doing so commonly invites chicken-and-egg issues between the
    :func:`init` and :func:`betse.lib.libs.reinit` methods and hence is
    inadvisable; in this case, however, the need to instantiate this singleton
    to display critical errors subsumes the need to instantiate this singleton
    in a more controlled manner.

    Parameters
    ----------
    title : str
        Title of this error to be displayed as the title of this message box.
    synopsis : str
        Synopsis of this error to be displayed as the text of this message box.
    exegesis : optional[str]
        Exegesis (i.e., explanation) of this error to be displayed as the
        so-called "informative text" of this message box below the synopsis of
        this error. Defaults to ``None``, in which case no such text is
        displayed.
    details : optional[str]
        Technical details of this error to be displayed as the so-called
        "detailed text" of this message box in monospaced font below both the
        synopsis and exegesis of this error in a discrete fold-down text area.
        Defaults to ``None``, in which case no such text is displayed.
    '''

    # Type check manually.
    assert isinstance(title,    str), '"{}" not a string.'.format(title)
    assert isinstance(synopsis, str), '"{}" not a string.'.format(synopsis)

    # Instantiate and initialize the "QApplication" singleton for this
    # application if needed *BEFORE* creating children widgets of this
    # singleton. Failure to do so typically results in non-human-readable
    # segmentation faults from Qt itself resembling:
    #
    #    QWidget: Must construct a QApplication before a QWidget
    #    [1]    30475 abort      betsee -v
    guiapp.init()

    # Message box displaying this error.
    error_box = QMessageBox()
    error_box.setWindowTitle(title)
    error_box.setText(synopsis)
    error_box.setIcon(QMessageBox.Critical)
    error_box.setStandardButtons(QMessageBox.Ok)

    # If this exception provides optional metadata, display this metadata.
    if exegesis is not None:
        assert isinstance(exegesis, str), '"{}" not a string.'.format(exegesis)
        error_box.setInformativeText(exegesis)
    if details is not None:
        assert isinstance(details, str), '"{}" not a string.'.format(details)
        error_box.setDetailedText(details)

    # Finalize this message box *AFTER* setting all widget proporties above.
    error_box.show()

    # Run this application's event loop, displaying this message box.
    error_box.exec_()


def show_exception(exception: Exception) -> None:
    '''
    Display the passed exception as a :mod:`QMessageBox`-driven modal message
    box in the current application widget, creating this widget if necessary.

    Parameters
    ----------
    exception : Exception
        Exception to be displayed.
    '''

    # Type check manually.
    assert isinstance(exception, Exception), (
        '"{}" not an exception.'.format(exception))

    # Human-readable synopsis and exegesis of this exception if defined (e.g.,
    # if this exception is an instance of the "BetseeException" superclass)
    # *OR* this exception message and None otherwise.
    exception_synopsis = getattr(exception, 'synopsis', str(exception))
    exception_exegesis = getattr(exception, 'exegesis', None)

    # If this exception has a human-readable title, use this title as is.
    if hasattr(exception, 'title'):
        exception_title = exception.title
    # Else, synthesize this title from metadata associated with this exception.
    else:
        # Class name of this exception.
        exception_classname = type(exception).__name__

        # Human-readable title of this exception, synthesized from this
        # machine-readable class name by inserting spaces between all
        # boundaries between non-capitalized and capitalized letters (e.g.,
        # from "ValueError" to "Value Error").
        exception_title = re.sub(
            r'([a-z])([A-Z])', r'\1 \2', exception_classname,)

    # Attempt to obtain an exception traceback via BETSE, which is *NOT*
    # guaranteed to exist at this point.
    try:
        from betse.util.io import ioexceptions
        _, exception_traceback = ioexceptions.get_metadata(exception)
    # If BETSE is unimportable, ignore this exception traceback.
    except ImportError:
        exception_traceback = None

    # Display this exception as a message box.
    show_error(
        title=exception_title,
        synopsis=exception_synopsis,
        exegesis=exception_exegesis,
        details=exception_traceback,
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
:mod:`PySide2`-based filetype functionality.
'''

# ....................{ IMPORTS                            }....................
# from PySide2.QtCore import QCoreApplication
from PySide2.QtGui import QImageReader
from betse.util.type.decorator.decmemo import func_cached
# from betse.util.io.log import logs
from betse.util.type.types import SetType  # type_check,

# ....................{ SELECTORS ~ read                   }....................
@func_cached
def get_image_read_filetypes() -> SetType:
    '''
    Set of all image filetypes readable by the low-level :class:`QImageReader`
    utility class internally required by all high-level image classes (e.g.,
    :class:`PySide2.QtGui.QImage`, :class:`PySide2.QtGui.QPicture`).

    For generality, these filetypes are *not* prefixed by a ``.`` delimiter.

    Examples
    ----------
        >>> from betsee.util.path import guifiletype
        >>> guifiletype.get_image_read_filetypes()
        ... {'bmp',
        ...  'cur',
        ...  'gif',
        ...  'ico',
        ...  'jpeg',
        ...  'jpg',
        ...  'pbm',
        ...  'pgm',
        ...  'png',
        ...  'ppm',
        ...  'svg',
        ...  'svgz',
        ...  'xbm',
        ...  'xpm'}
    '''

    # Avoid circular import dependencies.
    from betsee.util.type.text import guistr

    # Create, cache, and return the set of...
    return {
        # Decode this Qt-specific byte array into a Python string.
        guistr.decode_qbytearray_ascii(image_filetype_qbytearray)
        # For each "QByteArray" instance comprising an ASCII-encoded image
        # filetype readable by "QImageReader"...
        for image_filetype_qbytearray in QImageReader.supportedImageFormats()
    }



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
:class:`QLineEdit`-based simulation configuration widget subclasses.
'''

# ....................{ IMPORTS                           }....................
from PySide2.QtCore import QCoreApplication, Signal, Slot
from PySide2.QtWidgets import QLineEdit, QPushButton
from betse.exceptions import BetseMethodUnimplementedException
# from betse.util.io.log import logs
from betse.util.path import pathnames
from betse.util.type.types import type_check, StrOrNoneTypes
from betsee.gui.simconf.stack.widget.abc.guisimconfwdgeditscalar import (
    QBetseeSimConfEditScalarWidgetMixin)
from betsee.util.widget.abc.guiclipboardabc import (
    QBetseeClipboardScalarWidgetMixin)
from betsee.util.widget.stock.guilabel import QBetseeLabelImage

# ....................{ SUBCLASSES                        }....................
class QBetseeSimConfLineEdit(
    QBetseeClipboardScalarWidgetMixin,
    QBetseeSimConfEditScalarWidgetMixin,
    QLineEdit,
):
    '''
    Simulation configuration-specific line edit widget, interactively editing
    single-line strings backed by external simulation configuration files.
    '''

    # ..................{ SUPERCLASS ~ setter               }..................
    def setText(self, text_new: str) -> None:

        # Defer to the superclass setter.
        super().setText(text_new)

        # If this configuration is currently open, set the current value of
        # this simulation configuration alias to this widget's current value.
        self._set_alias_to_widget_value_if_sim_conf_open()

    # ..................{ MIXIN ~ property : read-only      }..................
    @property
    def undo_synopsis(self) -> str:
        return QCoreApplication.translate(
            'QBetseeSimConfLineEdit', 'edits to a text box')


    @property
    def _finalize_widget_change_signal(self) -> Signal:
        return self.editingFinished

    # ..................{ MIXIN ~ property : value          }..................
    @property
    def widget_value(self) -> str:
        return self.text()


    @widget_value.setter
    @type_check
    def widget_value(self, widget_value: str) -> None:

        # If this value is *NOT* a string, coerce this value into a string.
        # Since effectively all scalar values are safely coercable into strings
        # (due to their implementation of the special __str__() method), this
        # is guaranteed to be safe and hence need *NOT* be checked.
        if not isinstance(widget_value, str):
            widget_value = str(widget_value)

        # Set this widget's displayed value to the passed value by calling the
        # setText() method of our superclass rather than this subclass,
        # preventing infinite recursion. (See the superclass method docstring.)
        super().setText(widget_value)


    def _reset_widget_value(self) -> None:
        self.widget_value = ''

# ....................{ SUBCLASSES ~ pathname             }....................
class QBetseeSimConfPathnameLineEditABC(QBetseeSimConfLineEdit):
    '''
    Abstract base class of all simulation configuration-specific pathname line
    edit widget subclasses, interactively editing pathnames backed by external
    simulation configuration files.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def init(self, push_btn: QPushButton, *args, **kwargs) -> None:
        '''
        Finalize the initialization of this line edit, associated with all
        passed **sibling buddy widgets** (i.e., widgets spatially adjacent to
        this line edit, whose initialization is finalized by this method in a
        manner informing these widgets of their association to this line edit).

        Parameters
        ----------
        push_btn : QPushButton
            Push button "buddy" associated with this line edit. To display a
            path dialog selecting this pathname when this button is clicked,
            this method connects the :meth:`QPushButton.clicked` signal of this
            button to the :meth:`_set_text_to_pathname_selected` slot of this
            line edit. By convention, this button is typically labelled
            "Browse..." and situated to the right of this line edit.

        All remaining parameters are passed as is to the superclass
        :meth:`QBetseeSimConfEditScalarWidgetMixin.init` method.
        '''

        # Initialize our superclass with all remaining arguments.
        super().init(*args, **kwargs)

        # Connect this button's click signal to a slot displaying a path dialog
        # yielding the pathname to be displayed by this line edit.
        push_btn.clicked.connect(self._set_text_to_pathname_selected)

    # ..................{ SLOTS                             }..................
    @Slot()
    def _set_text_to_pathname_selected(self) -> None:
        '''
        Slot setting the text displayed by this line edit to the possibly
        non-existing pathname selected by a subclass-specific path dialog
        satisfying various constraints (e.g., image, subdirectory).

        This slot is connected to the :attr:`clicked` signal of the push button
        associated with this line edit at widget finalization time, for safety.
        '''

        # Pathname to initially display in the following dialog.
        old_pathname = self.text()

        # Pathname selected from a dialog displayed on clicking the push button
        # associated with this line edit if not cancelled *OR* "None" else.
        new_pathname = self._select_pathname(old_pathname)

        # If this dialog was *NOT* canceled, set the text displayed by this
        # line edit to this pathname.
        if new_pathname is not None:
            self.setText(new_pathname)

    # ..................{ SUBCLASS                          }..................
    # Subclasses are required to implement the following methods.

    @type_check
    def _select_pathname(self, init_pathname: str) -> StrOrNoneTypes:
        '''
        Possibly non-existing pathname interactively selected by the user on
        clicking the push button buddy associated with this line edit from a
        subclass-specific dialog displayed by this method if the user did not
        cancel this dialog *or* ``None`` otherwise (i.e., if the user cancelled
        this dialog).

        Parameters
        ----------
        init_pathname : str
            Pathname of the path to be initially displayed by this dialog
            (e.g., as externally specified by this simulation configuration).

        Returns
        ----------
        StrOrNoneTypes
            Either:
            * If this dialog was *not* cancelled, the possibly non-existing
              pathname selected from this dialog.
            * Else, ``None``.
        '''

        raise BetseMethodUnimplementedException()

# ....................{ SUBCLASSES ~ pathname : image     }....................
class QBetseeSimConfPathnameImageLineEdit(QBetseeSimConfPathnameLineEditABC):
    '''
    Simulation configuration-specific **image filename** (i.e., filenames with
    filetypes recognized by Pillow, the third-party image processing framework
    leveraged by BETSE itself) line edit widget, interactively editing image
    filenames backed by external simulation configuration files.

    Attributes
    ----------
    _image_label : QLabel
        Label "buddy" associated with this line edit. To preview the image
        whose filename is the text displayed by this line edit, this label's
        pixmap is read from this filename. By convention, this label is
        typically situated below this line edit.

    See Also
    ----------
    :func:`guifile.select_image_read`
        Further details.
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Nullify all instance variables for safety.
        self._image_label = None


    @type_check
    def init(self, image_label: QBetseeLabelImage, *args, **kwargs) -> None:
        '''
        Finalize the initialization of this line edit, associated with all
        passed **sibling buddy widgets** (i.e., widgets spatially adjacent to
        this line edit, whose initialization is finalized by this method in a
        manner informing these widgets of their association to this line edit).

        Parameters
        ----------
        image_label : QBetseeLabelImage
            Label "buddy" associated with this line edit. To preview the image
            whose filename is the text displayed by this line edit, this
            label's pixmap is read from this filename. By convention, this
            label is typically situated below this line edit. To preserve this
            image's aspect ratio, this widget is required to be an instance of
            the application-specific :class:`QBetseeLabelImage` class rather
            than application-agnostic :class:`QLabel` class, which does *not*
            preserve this image's aspect ratio.

        All remaining parameters are passed as is to the superclass
        :meth:`QBetseeSimConfPathnameLineEditABC.init` method.
        '''

        # Initialize our superclass with all remaining arguments.
        super().init(*args, **kwargs)

        # Classify all passed parameters.
        self._image_label = image_label

        # Finalize the initialization of this widget if needed.
        self._image_label.init_if_needed()

        # Instruct this label to auto-wrap long lines, required to display
        # non-fatal warnings subsequently added as this label's text in a
        # DPI-independent manner.
        self._image_label.setWordWrap(True)

    # ..................{ SUPERCLASS ~ getters              }..................
    # Called on opening and closing simulation configurations.
    def _get_widget_from_alias_value(self) -> object:

        # Value of the alias associated with this widget, equivalent to the
        # relative or absolute filename of expected image.
        image_filename = super()._get_widget_from_alias_value()

        # Preview an image only if a simulation configuration is open.
        self._image_label.setVisible(self._is_sim_open)

        # If a simulation configuration is open, preview this image.
        self._preview_image_if_sim_open(image_filename)

        # Return this value.
        return image_filename


    # Called on each user edit of this widget's value.
    def _get_alias_from_widget_value(self) -> object:

        # This widget's currently displayed value, equivalent to the relative
        # or absolute filename of the expected image.
        image_filename = super()._get_alias_from_widget_value()

        # If a simulation configuration is open, preview this image.
        self._preview_image_if_sim_open(image_filename)

        # Return this filename.
        return image_filename

    # ..................{ PREVIEWERS                        }..................
    #FIXME: Test this method with common edge cases, including:
    #
    #* A filename of filetype unrecognized by Pillow.
    #* A filename of filetype unrecognized by Qt.
    #* A file of recognized filetype that isn't actually an image.
    @type_check
    def _preview_image_if_sim_open(self, image_filename: str) -> None:
        '''
        Display a preview of the image with the passed filename as the pixmap
        of the label associated with this line edit if this image is
        previewable *or* log and/or display a non-fatal warning otherwise.

        If no simulation configuration is open, this method safely reduces to a
        noop without erroneously handling this image.

        Parameters
        ----------
        image_filename : str
            Absolute or relative filename of the image to be previewed.
        '''

        # Remove any prior content (e.g., image, text) displayed by this label,
        # regardless of whether subsequent logic displays any new such content.
        self._image_label.clear()

        # If no simulation configuration is open...
        if not self._is_sim_open:
            # Display a placeholder message.
            self._image_label.setText(QCoreApplication.translate(
                'QBetseeSimConfPathnameImageLineEdit', 'N/A'))

            # That's all she wrote.
            return
        # Else, a simulation configuration is open.

        # Absolute filename of this image. If relative, this filename is
        # converted into an absolute filename relative to the directory
        # containing this simulation configuration.
        image_filename_absolute = (
            image_filename if pathnames.is_absolute(image_filename) else
            pathnames.join(self._sim_conf.dirname, image_filename))

        # Load this image as the pixmap of the label buddied to this line edit.
        self._image_label.load_image(image_filename_absolute)

    # ..................{ SUPERCLASS ~ selectors            }..................
    @type_check
    def _select_pathname(self, init_pathname: str) -> StrOrNoneTypes:

        # Avoid circular import dependencies.
        from betsee.util.path import guifile

        # Return the absolute or relative filename of an existing selected
        # image of the parent directory containing the current simulation
        # configuration if this dialog was not canceled *OR* "None" otherwise.
        return guifile.select_image_read(
            init_pathname=init_pathname,
            parent_dirname=self._sim_conf.dirname)

# ....................{ SUBCLASSES ~ pathname : subdir     }....................
class QBetseeSimConfPathnameSubdirLineEdit(QBetseeSimConfPathnameLineEditABC):
    '''
    Simulation configuration-specific **subdirectory pathname** (i.e.,
    pathnames of subdirectories of top-level simulation configuration
    directories) line edit widget, interactively editing subdirectory pathnames
    backed by external simulation configuration files.

    For relocatability (i.e., to permit end users to trivially move simulation
    configurations to different directories), this line edit displays only the
    relative pathname of these subdirectories with respect to their parent
    directories; their absolute pathnames are *not* displayed.

    See Also
    ----------
    :func:`guidir.select_subdir`
        Further details.
    '''

    # ..................{ SUPERCLASS                        }..................
    @type_check
    def _select_pathname(self, init_pathname: str) -> StrOrNoneTypes:

        # Avoid circular import dependencies.
        from betsee.util.path import guidir

        # If the pathname of this subdirectory is absolute, raise an exception.
        pathnames.die_if_absolute(init_pathname)

        # Return the relative pathname of an existing selected subdirectory of
        # the parent directory containing the current simulation configuration
        # if this dialog was not canceled *OR* "None" otherwise.
        return guidir.select_subdir(
            init_pathname=init_pathname,
            parent_dirname=self._sim_conf.dirname)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level Qt-specific string handling functionality.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import Qt, QByteArray
# from betse.util.io.log import logs
# from betse.util.type import strs
from betse.util.type.text import mls
from betse.util.type.types import type_check

# ....................{ TESTERS                            }....................
@type_check
def is_rich(text: str) -> bool:
    '''
    ``True`` only if the passed text superficially appears to be HTML and thus
    satisfy Qt's definition of "rich text" rather than plaintext.

    Specifically:

    * If the :func:`Qt.mightBeRichText` function is available, this tester
      defers to that function.
    * Else, this tester falls back to the :func:`mls.is_ml` function.
    '''

    # Qt.mightBeRichText() function if this function exists or "None" otherwise.
    # Unfortunately, the shiboken2 Qt 5 bindings parser internally
    # leveraged by PySide2 occasionally fails to create valid bindings for a
    # variety of "PySide2.QtCore.Qt" functions -- including this function.
    mightBeRichText = getattr(Qt, 'mightBeRichText', None)

    # If this function exists, defer to this function as is.
    #
    # Else, this function does *NOT* exist. In this case, we defer to a
    # homegrown (albeit less reliable) solution detecting the existence of at
    # least one *ML (e.g., HTML) tag in this string.
    return mightBeRichText(text) if mightBeRichText else mls.is_ml(text)

# ....................{ DECODERS                           }....................
def decode_qbytearray_ascii(qbytearray: QByteArray) -> str:
    '''
    Decode the passed ASCII-encoded Qt-specific byte array into a Python string.
    '''

    # Python-specific byte array extracted from this Qt-specific byte array. If
    # the latter defines the more efficient constData() method, prefer that;
    # else, fall back to the less efficient data() method. For unknown reasons
    # (presumably relating to deficiencies in shiboken2-parsed PySide2
    # bindings), the data() method is *ALWAYS* defined but the constData()
    # method is only occasionally defined.
    byte_array = (
        qbytearray.constData() if hasattr(qbytearray, 'constData') else
        qbytearray.data())

    # Decode this Python byte array into a Python string.
    return str(byte_array, encoding='ascii')



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level support facilities for :mod:`pkg_resources`, a mandatory runtime
dependency simplifying inspection of application dependencies.

Caveats
----------
This submodule partially duplicates the existing
:mod:`betse.lib.setuptools.setuptool` submodule -- specifically, those
functions of that submodule required by the top-level installation-time
``setup.py`` script. Since BETSE is a third-party mandatory dependency of this
application, the :mod:`betse.lib.setuptools.setuptool` submodule provided by
that dependency is *not* necessarily importable from within this application's
``setup.py`` script.

This submodule thus violates the Don't Repeat Yourself (DRY) principle.
Unfortunately, doing so requires dropping support for type validation
implemented by the BETSE-specific :func:`betse.util.type.types.type_check`
decorator.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid race conditions during setuptools-based installation, this
# module may import *ONLY* from modules guaranteed to exist at the start of
# installation. This includes all standard Python and application modules but
# *NOT* third-party dependencies, which if currently uninstalled will only be
# installed at some later time in the installation.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from betsee.guiexception import BetseeLibException
from collections.abc import Mapping

# ....................{ CONVERTERS ~ dict-to-tuple        }....................
def convert_requirements_dict_to_tuple(requirements_dict: Mapping) -> tuple:
    '''
    Convert the passed dictionary of :mod:`setuptools`-specific requirements
    strings into a tuple of such strings.

    This dictionary is assumed to map from the :mod:`setuptools`-specific
    project name of a third-party dependency (e.g., ``NetworkX``) to the suffix
    of a :mod:`setuptools`-specific requirements string constraining this
    dependency (e.g., ``>= 1.11``). Each element of the resulting tuple is a
    string of the form `{key} {value}`, converted from a key-value pair of this
    dictionary in arbitrary order.

    Parameters
    ----------
    requirements_dict : MappingType
        Dictionary of :mod:`setuptools`-specific requirements strings in the
        format described above.

    Returns
    ----------
    tuple
        Tuple of :mod:`setuptools`-specific requirements strings in the format
        described above.
    '''

    return convert_requirements_dict_keys_to_tuple(
        requirements_dict, *requirements_dict.keys())


def convert_requirements_dict_keys_to_tuple(
    requirements_dict: Mapping, *requirement_names: str) -> tuple:
    '''
    Convert all key-value pairs of the passed dictionary of :mod:`setuptools`-
    specific requirements strings whose keys are the passed strings into a
    tuple of :mod:`setuptools`-specific requirements strings.

    Parameters
    ----------
    requirements_dict : MappingType
        Dictionary of requirements strings.
    requirement_names : Tuple[str]
        Tuple of keys identifying the key-value pairs of this dictionary to
        convert.

    Returns
    ----------
    tuple
        Tuple of :mod:`setuptools`-specific requirements strings in the above
        format.

    Raises
    ----------
    BetseeLibException
        If the passed key is *not* a key of this dictionary.

    See Also
    ----------
    :func:`convert_requirements_dict_to_tuple`
        Further details on the format of this dictionary and resulting strings.
    '''

    return tuple(
        convert_requirements_dict_key_to_str(
            requirements_dict, requirement_name)
        for requirement_name in requirement_names
    )


def convert_requirements_dict_key_to_str(
    requirements_dict: Mapping, requirement_name: str) -> str:
    '''
    Convert the key-value pair of the passed dictionary of :mod:`setuptools`-
    specific requirements strings whose key is the passed string into a
    :mod:`setuptools`-specific requirements string.

    Parameters
    ----------
    requirements_dict : MappingType
        Dictionary of requirements strings.
    requirement_names : str
        Key identifying the key-value pairs of this dictionary to convert.

    Returns
    ----------
    str
        Requirements string converted from this key-value pair.

    Raises
    ----------
    BetseeLibException
        If the passed key is *not* a key of this dictionary.

    See Also
    ----------
    :func:`convert_requirements_dict_to_tuple`
        Further details on the format of this dictionary and resulting string.
    '''

    # If this name is unrecognized, raise an exception.
    if requirement_name not in requirements_dict:
        raise BetseeLibException('Dependency "{}" unrecognized.'.format(
            requirement_name))

    # String constraining this requirement (e.g., ">= 3.14159265359", "[svg]")
    # if any or "None" otherwise.
    requirement_constraints = requirements_dict[requirement_name]

    # If this requirement is constrained, return the concatenation of this
    # requirement's name and constraints.
    if requirement_constraints:
        return '{} {}'.format(requirement_name, requirement_constraints)
    # Else, return only this requirement's name.
    else:
        return requirement_name
